/*  lcd.cpp
    Low level LCD driver for SED1520 122x32 graphic LCDs

    09.06.2001: tk, initial implementation.

    Copyright (c)2001 by Thomas Kindler, thomas.kindler@gmx.de

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of
    the License, or (at your option) any later version. Read the
    full License at http://www.gnu.org/copyleft for more details.
*/

#include <stdio.h>
#include <conio.h>
#include <string.h>
#include "lib/sc12/hwapi.h"
#include "lib/sc12/rtos.h"
#include "lib/helper/misc.h"
#include "lib/video/lcd.h"

// memory buffer for the LCD screen.
//
static unsigned char lcd_screen[LCD_HEIGHT/8][LCD_WIDTH];

/**
 * Send the screen buffer to the actual display memory
 * \note
 *   I tried REP_OUTSB, but some displays are too
 *   slow for this transfer.
 *
 */
void LCD_Update()
{
  for (int y=0; y<LCD_PAGES; y++)
  {
    if (!(inp(LCD_CHIP_CMD1)&&LCD_FLAG_BUSY))
    {
		printf("LCD-Busy!!!: %d\n",inp(LCD_CHIP_CMD1));
    }
    outp(LCD_CHIP_CMD2, LCD_SET_COLUMN);
    outp(LCD_CHIP_CMD2, LCD_SET_PAGE_0 + y);
    outp(LCD_CHIP_CMD1, LCD_SET_COLUMN);
    outp(LCD_CHIP_CMD1, LCD_SET_PAGE_0 + y);
    for (int x=0; x<(LCD_WIDTH/2); x++)
    {
      if (!(inp(LCD_CHIP_CMD2)&&LCD_FLAG_BUSY))
      {
        printf("LCD-Busy!!!: %d\n",inp(LCD_CHIP_CMD1));
      }
      outp(LCD_CHIP_DATA2, lcd_screen[y][x   ]);
      if (!(inp(LCD_CHIP_CMD1)&&LCD_FLAG_BUSY))
      {
        printf("LCD-Busy!!!: %d\n",inp(LCD_CHIP_CMD1));
      }
      outp(LCD_CHIP_DATA1, lcd_screen[y][x+LCD_WIDTH/2]);
    }
  }
}


/**
 * Set display brightness. This function MUST be called
 * during initialization, because it sets up the vertical
 * refresh timer for the LCD. (the brightness PWM signal
 * is derived from that clock)
 *
 * \param level  brigthness level 0..255
 *
 */
void LCD_SetBrightness(int level)
{
  long period = 2500000L/LCD_CLOCK;

  if (level <   0) level =   0;
  if (level > LCD_BRIGHTNESS_MAX) level = LCD_BRIGHTNESS_MAX;

  unsigned pwmOn, pwmOff;
  pwmOn  = (unsigned)(period*level/LCD_BRIGHTNESS_MAX);
  pwmOff = (unsigned)period - pwmOn;
  if (pwmOn  < LCD_MINPWMON )  pwmOn  = LCD_MINPWMON;
  if (pwmOff < LCD_MINPWMOFF)  pwmOff = LCD_MINPWMOFF;

  hal_init_timer(0, TIMER_CONT | TIMER_INT, pwmOn);
  hal_set_timer_duty(0, true, pwmOff);
  hal_start_timer(0);
}


/**
 * Get a pointer to the screen buffer
 */
char *LCD_GetScreenBuffer()
{
  return (char*)(&lcd_screen[0][0]);
}

/**
 * Clear the lcd_screen array.
 */
void LCD_Clear()
{
  memset(lcd_screen, 0x00, sizeof(lcd_screen));
}

/**
 * LCD_Init() must be called, before using any of the other
 * LCD functions. The display is initialized, and cleared.
 */
void LCD_Init()
{
  printf("initializing SED1520 LCD...\n");
  pfe_enable_bus(0xff, true);
  pfe_enable_pcs(LCD_CHIPSELECT_1);
  pfe_enable_pcs(LCD_CHIPSELECT_2);

  // enable refresh timer
  //
  pfe_enable_timer(TIMER_OUT0);
  LCD_SetBrightness(LCD_BRIGHTNESS_MAX);

  outp(LCD_CHIP_CMD1, LCD_RESET);  // reset
  outp(LCD_CHIP_CMD2, LCD_RESET);
  outp(LCD_CHIP_CMD1, LCD_STATIC_OFF);  // static drive off
  outp(LCD_CHIP_CMD2, LCD_STATIC_OFF);
  outp(LCD_CHIP_CMD1, LCD_DISPLAY_ON);  // display on
  outp(LCD_CHIP_CMD2, LCD_DISPLAY_ON);

  LCD_Clear();
  LCD_Update();
}


/**
 * Shut down the LCD controller.
 */
void LCD_Done()
{
  // enter power-save mode
  //
  outp(LCD_CHIP_CMD1, LCD_STATIC_ON);  // static drive on
  outp(LCD_CHIP_CMD2, LCD_STATIC_ON);
  outp(LCD_CHIP_CMD1, LCD_DISPLAY_OFF);  // display off
  outp(LCD_CHIP_CMD2, LCD_DISPLAY_OFF);

  // disable refresh timer
  //
  hal_stop_timer(0);

  // free pio pins
  //
  pfe_enable_pio(13, PIO_IPD);
  pfe_enable_pio( 2, PIO_IPU);
  pfe_enable_pio( 3, PIO_IPU); 
}

